@@ -27,14 +27,15 @@ class Api:
     you can specify which app and endpoint you wish to interact with.
 
     Valid attributes currently are:
+        * circuits
         * core (NetBox 3.5+)
         * dcim
-        * ipam
-        * circuits
-        * tenancy
         * extras
-        * virtualization
+        * ipam
+        * tenancy
         * users
+        * virtualization
+        * vpn (NetBox 3.7+)
         * wireless
 
     Calling any of these attributes will return
@@ -75,14 +76,15 @@ class Api:
         self.base_url = base_url
         self.http_session = requests.Session()
         self.threading = threading
+        self.circuits = App(self, "circuits")
         self.core = App(self, "core")
         self.dcim = App(self, "dcim")
-        self.ipam = App(self, "ipam")
-        self.circuits = App(self, "circuits")
-        self.tenancy = App(self, "tenancy")
         self.extras = App(self, "extras")
-        self.virtualization = App(self, "virtualization")
+        self.ipam = App(self, "ipam")
+        self.tenancy = App(self, "tenancy")
         self.users = App(self, "users")
+        self.virtualization = App(self, "virtualization")
+        self.vpn = App(self, "vpn")
         self.wireless = App(self, "wireless")
         self.plugins = PluginsApp(self)
 
