@@ -13,16 +13,13 @@ WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 """
-try:
-    import concurrent.futures as cf
-except ImportError:
-    pass
+import concurrent.futures as cf
 import json
-from six.moves.urllib.parse import urlencode
+from packaging import version
 
 
 def calc_pages(limit, count):
-    """ Calculate number of pages required for full results set. """
+    """Calculate number of pages required for full results set."""
     return int(count / limit) + (limit % count > 0)
 
 
@@ -43,18 +40,16 @@ class RequestError(Exception):
 
     """
 
-    def __init__(self, message):
-        req = message
-
+    def __init__(self, req):
         if req.status_code == 404:
-            message = "The requested url: {} could not be found.".format(req.url)
+            self.message = "The requested url: {} could not be found.".format(req.url)
         else:
             try:
-                message = "The request failed with code {} {}: {}".format(
+                self.message = "The request failed with code {} {}: {}".format(
                     req.status_code, req.reason, req.json()
                 )
             except ValueError:
-                message = (
+                self.message = (
                     "The request failed with code {} {} but more specific "
                     "details were not returned in json. Check the NetBox Logs "
                     "or investigate this exception's error attribute.".format(
@@ -62,30 +57,32 @@ class RequestError(Exception):
                     )
                 )
 
-        super(RequestError, self).__init__(message)
+        super().__init__(self.message)
         self.req = req
         self.request_body = req.request.body
         self.base = req.url
         self.error = req.text
 
+    def __str__(self):
+        return self.message
+
 
 class AllocationError(Exception):
     """Allocation Exception
 
     Used with available-ips/available-prefixes when there is no
-    room for allocation and NetBox returns 204 No Content.
+    room for allocation and NetBox returns 409 Conflict.
     """
 
-    def __init__(self, message):
-        req = message
-
-        message = "The requested allocation could not be fulfilled."
-
-        super(AllocationError, self).__init__(message)
+    def __init__(self, req):
+        super().__init__(req)
         self.req = req
         self.request_body = req.request.body
         self.base = req.url
-        self.error = message
+        self.error = "The requested allocation could not be fulfilled."
+
+    def __str__(self):
+        return self.error
 
 
 class ContentError(Exception):
@@ -96,21 +93,20 @@ class ContentError(Exception):
     exception is raised in those cases.
     """
 
-    def __init__(self, message):
-        req = message
-
-        message = (
-            "The server returned invalid (non-json) data. Maybe not " "a NetBox server?"
-        )
-
-        super(ContentError, self).__init__(message)
+    def __init__(self, req):
+        super().__init__(req)
         self.req = req
         self.request_body = req.request.body
         self.base = req.url
-        self.error = message
+        self.error = (
+            "The server returned invalid (non-json) data. Maybe not a NetBox server?"
+        )
+
+    def __str__(self):
+        return self.error
 
 
-class Request(object):
+class Request:
     """Creates requests to the Netbox API
 
     Responsible for building the url and making the HTTP(S) requests to
@@ -121,8 +117,6 @@ class Request(object):
         correlate to the filters a given endpoint accepts.
         In (e.g. /api/dcim/devices/?name='test') 'name': 'test'
         would be in the filters dict.
-    :param private_key: (str, optional) The user's private key as a
-        string.
     """
 
     def __init__(
@@ -130,10 +124,10 @@ class Request(object):
         base,
         http_session,
         filters=None,
+        limit=None,
+        offset=None,
         key=None,
         token=None,
-        private_key=None,
-        session_key=None,
         threading=False,
     ):
         """
@@ -146,35 +140,43 @@ class Request(object):
                 In (e.g. /api/dcim/devices/?name='test') 'name': 'test'
                 would be in the filters dict.
             key (int, optional): database id of the item being queried.
-            private_key (string, optional): The user's private key as a
-                string.
         """
         self.base = self.normalize_url(base)
-        self.filters = filters
+        self.filters = filters or None
         self.key = key
         self.token = token
-        self.private_key = private_key
-        self.session_key = session_key
         self.http_session = http_session
         self.url = self.base if not key else "{}{}/".format(self.base, key)
         self.threading = threading
+        self.limit = limit
+        self.offset = offset
 
     def get_openapi(self):
-        """ Gets the OpenAPI Spec """
+        """Gets the OpenAPI Spec"""
         headers = {
-            "Content-Type": "application/json;",
+            "Accept": "application/json",
+            "Content-Type": "application/json",
         }
-        req = self.http_session.get(
-            "{}docs/?format=openapi".format(self.normalize_url(self.base)),
-            headers=headers,
-        )
+
+        current_version = version.parse(self.get_version())
+        if current_version >= version.parse("3.5"):
+            req = self.http_session.get(
+                "{}schema/".format(self.normalize_url(self.base)),
+                headers=headers,
+            )
+        else:
+            req = self.http_session.get(
+                "{}docs/?format=openapi".format(self.normalize_url(self.base)),
+                headers=headers,
+            )
+
         if req.ok:
             return req.json()
         else:
             raise RequestError(req)
 
     def get_version(self):
-        """ Gets the API version of NetBox.
+        """Gets the API version of NetBox.
 
         Issues a GET request to the base URL to read the API version from the
         response headers.
@@ -184,50 +186,29 @@ class Request(object):
         present in the headers.
         """
         headers = {
-            "Content-Type": "application/json;",
+            "Content-Type": "application/json",
         }
-        req = self.http_session.get(self.normalize_url(self.base), headers=headers,)
+        req = self.http_session.get(
+            self.normalize_url(self.base),
+            headers=headers,
+        )
         if req.ok:
             return req.headers.get("API-Version", "")
         else:
             raise RequestError(req)
 
-    def get_session_key(self):
-        """Requests session key
-
-        Issues a GET request to the `get-session-key` endpoint for
-        subsequent use in requests from the `secrets` endpoint.
-
-        :Returns: String containing session key.
-        """
-        req = self.http_session.post(
-            "{}secrets/get-session-key/?preserve_key=True".format(self.base),
-            headers={
-                "accept": "application/json",
-                "authorization": "Token {}".format(self.token),
-                "Content-Type": "application/x-www-form-urlencoded",
-            },
-            data=urlencode({"private_key": self.private_key.strip("\n")}),
-        )
-        if req.ok:
-            try:
-                return req.json()["session_key"]
-            except json.JSONDecodeError:
-                raise ContentError(req)
-        else:
-            raise RequestError(req)
-
     def get_status(self):
-        """ Gets the status from /api/status/ endpoint in NetBox.
+        """Gets the status from /api/status/ endpoint in NetBox.
 
         :Returns: Dictionary as returned by NetBox.
         :Raises: RequestError if request is not successful.
         """
-        headers = {"Content-Type": "application/json;"}
+        headers = {"Content-Type": "application/json"}
         if self.token:
             headers["authorization"] = "Token {}".format(self.token)
         req = self.http_session.get(
-            "{}status/".format(self.normalize_url(self.base)), headers=headers,
+            "{}status/".format(self.normalize_url(self.base)),
+            headers=headers,
         )
         if req.ok:
             return req.json()
@@ -235,23 +216,20 @@ class Request(object):
             raise RequestError(req)
 
     def normalize_url(self, url):
-        """ Builds a url for POST actions.
-        """
+        """Builds a url for POST actions."""
         if url[-1] != "/":
             return "{}/".format(url)
 
         return url
 
     def _make_call(self, verb="get", url_override=None, add_params=None, data=None):
-        if verb in ("post", "put"):
-            headers = {"Content-Type": "application/json;"}
+        if verb in ("post", "put") or verb == "delete" and data:
+            headers = {"Content-Type": "application/json"}
         else:
-            headers = {"accept": "application/json;"}
+            headers = {"accept": "application/json"}
 
         if self.token:
             headers["authorization"] = "Token {}".format(self.token)
-        if self.session_key:
-            headers["X-Session-Key"] = self.session_key
 
         params = {}
         if not url_override:
@@ -264,7 +242,7 @@ class Request(object):
             url_override or self.url, headers=headers, params=params, json=data
         )
 
-        if req.status_code == 204 and verb == "post":
+        if req.status_code == 409 and verb == "post":
             raise AllocationError(req)
         if verb == "delete":
             if req.ok:
@@ -305,36 +283,19 @@ class Request(object):
             endpoint.
         """
 
-        def req_all():
-            req = self._make_call(add_params=add_params)
-            if isinstance(req, dict) and req.get("results") is not None:
-                ret = req["results"]
-                first_run = True
-                while req["next"]:
-                    # Not worrying about making sure add_params kwargs is
-                    # passed in here because results from detail routes aren't
-                    # paginated, thus far.
-                    if first_run:
-                        req = self._make_call(
-                            add_params={
-                                "limit": req["count"],
-                                "offset": len(req["results"]),
-                            }
-                        )
-                    else:
-                        req = self._make_call(url_override=req["next"])
-                    first_run = False
-                    ret.extend(req["results"])
-                return ret
-            else:
-                return req
-
-        def req_all_threaded(add_params):
-            if add_params is None:
-                # Limit must be 0 to discover the max page size
-                add_params = {"limit": 0}
-            req = self._make_call(add_params=add_params)
-            if isinstance(req, dict) and req.get("results") is not None:
+        if not add_params and self.limit is not None:
+            add_params = {"limit": self.limit}
+            if self.limit and self.offset is not None:
+                # if non-zero limit and some offset -> add offset
+                add_params["offset"] = self.offset
+        req = self._make_call(add_params=add_params)
+        if isinstance(req, dict) and req.get("results") is not None:
+            self.count = req["count"]
+            if self.offset is not None:
+                # only yield requested page results if paginating
+                for i in req["results"]:
+                    yield i
+            elif self.threading:
                 ret = req["results"]
                 if req.get("next"):
                     page_size = len(req["results"])
@@ -347,21 +308,40 @@ class Request(object):
                         ret.extend(req["results"])
                     else:
                         self.concurrent_get(ret, page_size, page_offsets)
-
-                return ret
+                for i in ret:
+                    yield i
             else:
-                return req
-
-        if self.threading:
-            return req_all_threaded(add_params)
-
-        return req_all()
+                first_run = True
+                for i in req["results"]:
+                    yield i
+                while req["next"]:
+                    # Not worrying about making sure add_params kwargs is
+                    # passed in here because results from detail routes aren't
+                    # paginated, thus far.
+                    if first_run:
+                        req = self._make_call(
+                            add_params={
+                                "limit": self.limit or req["count"],
+                                "offset": len(req["results"]),
+                            }
+                        )
+                    else:
+                        req = self._make_call(url_override=req["next"])
+                    first_run = False
+                    for i in req["results"]:
+                        yield i
+        elif isinstance(req, list):
+            self.count = len(req)
+            for i in req:
+                yield i
+        else:
+            self.count = len(req)
+            yield req
 
     def put(self, data):
         """Makes PUT request.
 
-        Makes a PUT request to NetBox's API. Adds the session key to
-        headers if the `private_key` attribute was populated.
+        Makes a PUT request to NetBox's API.
 
         :param data: (dict) Contains a dict that will be turned into a
             json object and sent to the API.
@@ -374,13 +354,12 @@ class Request(object):
     def post(self, data):
         """Makes POST request.
 
-        Makes a POST request to NetBox's API. Adds the session key to
-        headers if the `private_key` attribute was populated.
+        Makes a POST request to NetBox's API.
 
         :param data: (dict) Contains a dict that will be turned into a
             json object and sent to the API.
         :raises: RequestError if req.ok returns false.
-        :raises: AllocationError if req.status_code is 204 (No Content)
+        :raises: AllocationError if req.status_code is 409 (Conflict)
             as with available-ips and available-prefixes when there is
             no room for the requested allocation.
         :raises: ContentError if response is not json.
@@ -388,18 +367,20 @@ class Request(object):
         """
         return self._make_call(verb="post", data=data)
 
-    def delete(self):
+    def delete(self, data=None):
         """Makes DELETE request.
 
         Makes a DELETE request to NetBox's API.
 
+        :param data: (list) Contains a dict that will be turned into a
+            json object and sent to the API.
         Returns:
             True if successful.
 
         Raises:
             RequestError if req.ok doesn't return True.
         """
-        return self._make_call(verb="delete")
+        return self._make_call(verb="delete", data=data)
 
     def patch(self, data):
         """Makes PATCH request.
@@ -438,4 +419,6 @@ class Request(object):
         :returns: Int of number of objects query returned.
         """
 
-        return self._make_call(add_params={"limit": 1})["count"]
+        if not hasattr(self, "count"):
+            self.count = self._make_call(add_params={"limit": 1, "brief": 1})["count"]
+        return self.count
